/*
 *	Auxiliary functions for processing PDF files
 *
 *	(c) 2018 Martin Mares <mj@ucw.cz>
 */

#ifndef _PDF_TOOLS_H
#define _PDF_TOOLS_H

#include <cmath>
#include <string>

#include <qpdf/QPDF.hh>

/*** Basic macros and constants ***/

#define A4_WIDTH 595
#define A4_HEIGHT 842
static const double mm = 72/25.4;

/*** Transformation matrices ***/

struct pdf_matrix {
	/*
	 *  A transformation matrix corresponds to the linear transform
	 *
	 *            (a b 0)
	 *  (x y 1) * (c d 0) = (ax+cy+e bx+dy+f 1)
	 *            (e f 1)
	 *
	 *  We represent the non-trivial coefficients of the matrix by
	 *  an array {a,b,c,d,e,f}.
	 */
	double m[6];

	pdf_matrix() {
		m[0] = 1;
		m[1] = 0;
		m[2] = 0;
		m[3] = 1;
		m[4] = m[5] = 0;
	}

	pdf_matrix(double a, double b, double c, double d, double e, double f)
	{
		m[0] = a;
		m[1] = b;
		m[2] = c;
		m[3] = d;
		m[4] = e;
		m[5] = f;
	}

	void apply(double *px, double *py)
	{
		double x=*px, y=*py;
		*px = m[0]*x + m[2]*y + m[4];
		*py = m[1]*x + m[3]*y + m[5];
	}

	// A*B is a matrix which transforms first by A and then by B
	pdf_matrix operator *(pdf_matrix y)
	{
		return pdf_matrix(
			m[0]*y.m[0] + m[1]*y.m[2],
			m[0]*y.m[1] + m[1]*y.m[3],
			m[2]*y.m[0] + m[3]*y.m[2],
			m[2]*y.m[1] + m[3]*y.m[3],
			m[4]*y.m[0] + m[5]*y.m[2] + y.m[4],
			m[4]*y.m[1] + m[5]*y.m[3] + y.m[5]
		);
	}

	void concat(pdf_matrix y)
	{
		pdf_matrix t = *this * y;
		for (int i=0; i<6; i++)
			m[i] = t.m[i];
	}

	void shift(double dx, double dy)
	{
		m[4] += dx;
		m[5] += dy;
	}

	void scale(double s)
	{
		concat(pdf_matrix(s, 0, 0, s, 0, 0));
	}

	void scale(double sx, double sy)
	{
		concat(pdf_matrix(sx, 0, 0, sy, 0, 0));
	}

	void rotate_rad(double angle)
	{
		double c = std::cos(angle), s = std::sin(angle);
		concat(pdf_matrix(c, s, -s, c, 0, 0));
	}

	void rotate_deg(double angle)
	{
		rotate_rad(angle/180. * M_PI);
	}

	std::string to_string();
};

/*** Bounding boxes ***/

struct BBox {
	double x_min, x_max, y_min, y_max;
	BBox() {
		x_min = y_min = x_max = y_max = 0;
	}
	BBox(double xmin, double ymin, double xmax, double ymax) {
		x_min = xmin, x_max = xmax;
		y_min = ymin, y_max = ymax;
	}
	BBox(double x, double y) {
		x_min = 0, x_max = x;
		y_min = 0, y_max = y;
	}
	BBox(QPDFObjectHandle box);
	QPDFObjectHandle to_array();
	string to_rect();
	double width() { return x_max - x_min; }
	double height() { return y_max - y_min; }
	void transform(pdf_matrix &m);
	BBox transformed(pdf_matrix &m);
	void enlarge(double by);
	BBox enlarged(double by);
	void join(BBox &with);
	void intersect(BBox &with);
private:
	bool parse(QPDFObjectHandle h);
};

/*** Miscellaneous ***/

QPDFObjectHandle unicode_string(std::string s);
QPDFObjectHandle page_to_xobject(QPDF *out, QPDFObjectHandle page);
string pdf_coord(double x, uint digits=1);
pdf_matrix pdf_rotation_matrix(int deg, double width, double height);

#endif
