/*
 *	PaperJam -- Low-level handling of PDFs
 *
 *	(c) 2018--2022 Martin Mares <mj@ucw.cz>
 */

#include <cassert>
#include <cstdlib>
#include <cstdio>
#include <cstring>
#include <unistd.h>
#include <sys/wait.h>

#include "jam.h"

#include <qpdf/QPDFWriter.hh>

static QPDF in_pdf;
static QPDF out_pdf;

static void do_recalc_bbox(vector<page *> &pages, const char *in_name);

string out_context::new_resource(const string type)
{
  return "/" + type + to_string(++res_cnt);
}

class in_page : public page {
  QPDFObjectHandle pdf_page;
  QPDFObjectHandle xobject;
public:
  BBox media_box;
  void render(out_context *out, pdf_matrix xform);
  void debug_dump() { debug("Input page %d", index); }
  in_page(QPDFObjectHandle inpg, int idx);
  int get_rotate();
};

in_page::in_page(QPDFObjectHandle inpg, int idx)
{
  pdf_page = inpg;
  xobject = QPDFObjectHandle::newNull();
  index = idx;

  media_box = BBox(inpg.getKey("/MediaBox"));
  width = media_box.width();
  height = media_box.height();

  QPDFObjectHandle art_box = inpg.getKey("/ArtBox");
  if (art_box.isNull())
    art_box = inpg.getKey("/CropBox");
  if (art_box.isNull())
    image_box = BBox(width, height);
  else
    {
      image_box = BBox(art_box);
      image_box.x_min -= media_box.x_min;
      image_box.x_max -= media_box.x_min;
      image_box.y_min -= media_box.y_min;
      image_box.y_max -= media_box.y_min;
    }
}

void in_page::render(out_context *out, pdf_matrix xform)
{
  // Convert page to xobject
  if (xobject.isNull())
    xobject = out->pdf->makeIndirectObject( page_to_xobject(out->pdf, out->pdf->copyForeignObject(pdf_page)) );
  string xobj_res = out->new_resource("XO");
  out->xobjects.replaceKey(xobj_res, xobject);

  pdf_matrix m;
  m.shift(-media_box.x_min, -media_box.y_min);
  m.concat(xform);

  out->contents += "q " + m.to_string() + " cm " + xobj_res + " Do Q ";
}

int in_page::get_rotate()
{
  QPDFObjectHandle rotate = pdf_page.getKey("/Rotate");
  if (rotate.isNull())
    return 0;
  else if (rotate.isInteger())
    {
      long long deg = rotate.getIntValue();
      if (deg < 0 || deg >= 360 || deg % 90)
	{
	  warn("Page #%d: /Rotate must be 0, 90, 180 or 270", index);
	  return 0;
	}
      else
	return deg;
    }
  else
    {
      warn("Page #%d: /Rotate is not an integer", index);
      return 0;
    }
}

void debug_pages(vector<page *> &pages)
{
  if (!debug_level)
    return;

  for (auto pg: pages)
    {
      debug("Page #%d: media[%.3f %.3f] image[%.3f %.3f %.3f %.3f][%.3f %.3f]",
	pg->index,
	pg->width, pg->height,
	pg->image_box.x_min, pg->image_box.y_min, pg->image_box.x_max, pg->image_box.y_max,
	pg->image_box.width(), pg->image_box.height());
      if (debug_level > 2)
	{
	  debug_indent += 4;
	  pg->debug_dump();
	  debug_indent -= 4;
	}
    }
}

static vector<page *> apply_input_xforms(vector<page *> in_pages)
{
  vector<page *> out_pages;

  for (auto pg: in_pages)
    {
      in_page * in_pg = dynamic_cast<in_page *>(pg);
      if (in_pg)
	{
	  int deg = in_pg->get_rotate();
	  if (deg)
	    pg = new xform_page(pg, "/Rotate", pdf_rotation_matrix(deg, pg->width, pg->height));
	}
      out_pages.push_back(pg);
    }

  return out_pages;
}

vector<page *> run_command_list(list<cmd *> &cmds, vector<page *> &pages)
{
  debug("# Input");
  debug_pages(pages);

  for (auto c: cmds)
    {
      debug("# Executing %s", c->def->name);
      debug_indent += 4;
      try
	{
	  pages = c->exec->process(pages);
	}
      catch (exception &e)
	{
	  die("Error in %s: %s", c->def->name, e.what());
	}
      debug_indent -= 4;
      debug_pages(pages);
    }

  return pages;
}

static void make_info_dict()
{
  // Create info dictionary if it did not exist yet
  QPDFObjectHandle trailer = out_pdf.getTrailer();
  QPDFObjectHandle info = trailer.getKey("/Info");
  if (info.isNull())
    {
      info = QPDFObjectHandle::newDictionary();
      trailer.replaceKey("/Info", info);
    }
  else
    assert(info.isDictionary());

  info.replaceKey("/Producer", unicode_string("PaperJam"));

  // Copy entries from the source file's info dictionary
  QPDFObjectHandle orig_trailer = in_pdf.getTrailer();
  QPDFObjectHandle orig_info = orig_trailer.getKey("/Info");
  if (!orig_info.isNull())
    {
      const string to_copy[] = { "/Title", "/Author", "/Subject", "/Keywords", "/Creator", "/CreationDate" };
      for (string key: to_copy)
	info.replaceOrRemoveKey(key, orig_info.getKey(key));
    }
}

void process(list<cmd *> &cmds)
{
  debug("### Reading input");
  in_pdf.processFile(in_name);
  in_pdf.pushInheritedAttributesToPage();
  out_pdf.emptyPDF();

  vector<QPDFObjectHandle> const &in_pages = in_pdf.getAllPages();
  vector<page *> pages;

  QPDFObjectHandle page_copy = out_pdf.copyForeignObject(in_pages[0]);

  int cnt = 0;
  for (auto inpg: in_pages)
    pages.push_back(new in_page(inpg, ++cnt));

  if (recalc_bbox)
    do_recalc_bbox(pages, in_name);

  if (!no_auto_transforms)
    {
      debug("### Applying input transforms");
      pages = apply_input_xforms(pages);
    }

  debug("### Running commands");
  pages = run_command_list(cmds, pages);

  debug("### Writing output");
  int out_page = 0;
  for (auto pg: pages)
    {
      ++out_page;
      if (debug_level > 1)
	{
	  debug("Page #%d", out_page);
	  debug_indent += 4;
	  pg->debug_dump();
	  debug_indent -= 4;
	}

      out_context out;
      out.pdf = &out_pdf;
      out.resources = QPDFObjectHandle::newDictionary();
      out.resources.replaceKey("/ProcSet", QPDFObjectHandle::parse("[/PDF]"));
      out.xobjects = QPDFObjectHandle::newDictionary();
      out.egstates = QPDFObjectHandle::newDictionary();
      pg->render(&out, pdf_matrix());

      QPDFObjectHandle contents = QPDFObjectHandle::newStream(&out_pdf, out.contents);

      // Create the page object
      QPDFObjectHandle out_page = out_pdf.makeIndirectObject(QPDFObjectHandle::newDictionary());
      out_page.replaceKey("/Type", QPDFObjectHandle::newName("/Page"));
      out_page.replaceKey("/MediaBox", BBox(pg->width, pg->height).to_array());
      // FIXME:
      // out_page.replaceKey("/CropBox", pg->image_box.to_array());
      out_page.replaceKey("/Contents", contents);
      if (!out.xobjects.getKeys().empty())
	out.resources.replaceKey("/XObject", out.xobjects);
      if (!out.egstates.getKeys().empty())
	out.resources.replaceKey("/ExtGState", out.egstates);
      out_page.replaceKey("/Resources", out.resources);
      out_pdf.addPage(out_page, false);
    }

  // Produce info dictionary
  make_info_dict();

  // Write the output file
  QPDFWriter writer(out_pdf, out_name);
  writer.write();
  debug("### Done");
}

/*** Re-calculation of bboxes ***/

vector<BBox> gs_bboxes(const char *in)
{
  int pipes[2];
  if (pipe(pipes) < 0)
    die("Cannot create pipe: %m");

  pid_t pid = fork();
  if (pid < 0)
    die("Cannot fork: %m");

  if (!pid)
    {
      close(pipes[0]);
      dup2(pipes[1], 1);
      dup2(pipes[1], 2);
      close(pipes[1]);
      execlp("gs", "gs", "-sDEVICE=bbox", "-dSAFER", "-dBATCH", "-dNOPAUSE", "-q", in, NULL);
      die("Cannot execute gs: %m");
    }

  close(pipes[1]);
  FILE *f = fdopen(pipes[0], "r");
  if (!f)
    die("fdopen failed: %m");

  char line[1024];
  vector<BBox> bboxes;
  while (fgets(line, sizeof(line), f))
    {
      char *eol = strchr(line, '\n');
      if (!eol)
	die("Ghostscript produced too long lines");
      *eol = 0;

      if (!strncmp(line, "%%HiResBoundingBox: ", 20))
	{
	  double x1, y1, x2, y2;
	  if (sscanf(line+20, "%lf%lf%lf%lf", &x1, &y1, &x2, &y2) != 4)
	    die("Cannot parse Ghostscript output: %s", line);
	  bboxes.push_back(BBox(x1, y1, x2, y2));
	}
      else if (line[0] != '%')
	fprintf(stderr, "%s\n", line);
    }
  fclose(f);

  int stat;
  if (waitpid(pid, &stat, 0) < 0)
    die("wait failed: %m");
  if (!WIFEXITED(stat) || WEXITSTATUS(stat))
    die("Ghostscript failed");

  return bboxes;
}

static void do_recalc_bbox(vector<page *> &pages, const char *in_name)
{
  debug("Calling Ghostscript to re-calculate bounding boxes");
  vector<BBox> bboxes = gs_bboxes(in_name);
  if (pages.size() != bboxes.size())
    die("Ghostscript failed to produce the right number of bboxes");

  for (size_t i=0; i<pages.size(); i++)
    pages[i]->image_box = bboxes[i];
}

// Transformed page

xform_page::xform_page(page *p, const char *desc, pdf_matrix xf)
{
  orig_page = p;
  index = p->index;
  description = desc;
  xform = xf;

  BBox media(p->width, p->height);
  media.transform(xf);
  width = media.width();
  height = media.height();

  image_box = p->image_box;
  image_box.transform(xf);
}

void xform_page::debug_dump()
{
  debug("Transform (%s): [%s]", description, xform.to_string().c_str());
  orig_page->debug_dump();
}

void xform_page::render(out_context *out, pdf_matrix parent_xform)
{
  orig_page->render(out, xform * parent_xform);
}
